/*----------------------------------------------------------------------
  File    : cli.c
  Contents: probabilistic and fuzzy cluster induction
  Author  : Christian Borgelt
  History : 2001.09.15 file created from file mlpt.c
            2002.09.09 neural network update methods added
            2003.01.30 data normalization moved to cluster.c
            2003.03.20 bug in function msfnpar fixed
            2003.05.15 options -n, -g, -z changed, options -q, -j added
            2003.05.16 noise clustering added (option -y)
            2003.06.07 cluster size scaling factor added
            2003.08.16 slight changes in error message output
            2004.02.25 source files cli.c amd mcli.c combined
            2004.03.18 normalization parameters added
            2004.03.19 cluster weight regularization added
            2004.04.06 parameters for competitive learning added
            2004.07.13 normalization of center vectors added
            2004.07.27 option -C added (epochs to update centers only)
            2007.01.10 adapted to changed cluster interface
            2007.02.14 adapted to changed module tabread
            2007.03.16 table and matrix version combined (option -M)
            2007.07.07 bug with excluded target fixed (AM_MARKED)
            2007.10.10 evaluation of attribute directions added
            2008.03.01 redesign for feature weighting completed
            2008.04.14 option -K added (neural network update)
            2008.04.17 option -P added (for faster experiments)
            2008.04.22 sorting the clusters by their centers added
            2011.12.16 adapted to modified attset and utility modules
            2012.05.02 adapted to modified attset and utility modules
            2013.08.16 adapted to definitions ATTID, VALID, DIMID etc.
            2013.08.29 adapted to new function as_target()
            2014.10.24 changed from LGPL license to MIT license
            2016.02.25 iterative cluster set reduction added
            2016.02.26 maximum change for termination DBL_EPSILON
            2018.06.10 more initialization methods added (e.g. kmeans++)
            2018.07.17 (bounds-based) exact hard c-means methods added
            2018.07.25 trimmed maximin init. added (cut top quantile)
            2018.08.18 extended functions (attset/table) removed
            2018.08.27 annular algorithm [Drake 2013] added
            2018.08.28 yinyang algorithm [Ding et al. 2015] added
            2018.08.29 stratum algorithm (modified annular) added
            2018.08.31 shallot algorithm (improved exponion) added
            2018.09.01 coordinate grid (stratum) algorithm added
            2019.07.09 distribution based initialization added to plain
            2019.07.10 non-squared distance based kmeans++ init. added
            2019.11.15 adapted to changed interface of mat_shuffle()
            2022.05.10 no scaling if no scales in cluster input file
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#include "arrays.h"
#include "random.h"
#include "params.h"
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#ifndef AS_DESC
#define AS_DESC
#endif
#include "attset.h"
#include "attmap.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#ifndef TAB_MATRIX
#define TAB_MATRIX
#endif
#include "table.h"
#ifndef CLS_PARSE
#define CLS_PARSE
#endif
#include "cluster.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "cli"
#define DESCRIPTION "probabilistic and fuzzy cluster induction"
#define VERSION     "version 5.7 (2022.05.10)         " \
                    "(c) 2001-2022   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse error */
#define E_TPLCNT    (-10)       /* data file is empty */
#define E_ATTCNT    (-11)       /* no usable attributes found */
#define E_INPCNT    (-12)       /* no usable inputs found */
#define E_UNKTRG    (-13)       /* missing target attribute */
#define E_MULTRG    (-14)       /* multiple target attributes */
#define E_EXCNT     (-15)       /* invalid number of excluded columns */
#define E_MODE      (-16)       /* invalid init. or norm. mode */
#define E_INIPAR    (-17)       /* invalid initialization parameter */
#define E_OUTIND    (-18)       /* invalid outlier indicator */
#define E_OUTID     (-19)       /* invalid outlier identifier/index */
#define E_OIDCNT    (-20)       /* invalid number of outlier inds. */
#define E_RFNPAR    (-21)       /* invalid radial function parameter */
#define E_NRMPAR    (-22)       /* invalid normalization parameter */
#define E_MFEXP     (-23)       /* invalid membership exponent */
#define E_METHOD    (-24)       /* invalid update method */
#define E_MODIFY    (-25)       /* invalid update modifier */
#define E_MOMENT    (-26)       /* invalid momentum coefficient */
#define E_RADIUS    (-27)       /* invalid initial radius */
#define E_REGPAR    (-28)       /* invalid regularization parameter */
#define E_LRATE     (-29)       /* invalid learning rate */
#define E_UPDPAR    (-30)       /* invalid update parameter */
#define E_EPOCHS    (-31)       /* invalid number of epochs */

#ifdef BENCH                    /* if benchmark version */
#define CLK_FMT     ".6f"       /* print time precisely */
#else                           /* if standard version */
#define CLK_FMT     ".2f"       /* print time less precisely */
#endif                          /* definitions for time measurements */
#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*--------------------------------------------------------------------*/
#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

#if DIMID!=ATTID
#error "DIMID and ATTID must be the same data type"
#endif

#if DIMID!=TPLID
#error "DIMID and TPLID must be the same data type"
#endif

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- mode information --- */
  int  code;                    /* code        of mode */
  char *name;                   /* name        of mode */
  char *desc;                   /* description of mode */
} MODEINFO;                     /* (mode information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_TPLCNT  -10 */  "data file is empty (no data points)",
  /* E_ATTCNT  -11 */  "no (usable) attributes (need at least 1)",
  /* E_INPCNT  -12 */  "no (usable) inputs (need at least 1)",
  /* E_UNKTRG  -13 */  "unknown target attribute '%s'",
  /* E_MULTRG  -14 */  "multiple target attributes found",
  /* E_EXCNT   -15 */  "invalid number of excluded columns %s",
  /* E_MODE    -16 */  "unknown mode/method %s",
  /* E_INIPAR  -17 */  "invalid initialization parameter %g",
  /* E_OUTIND  -18 */  "invalid outlier indicator %s\n",
  /* E_OUTID   -19 */  "invalid outlier index %g\n",
  /* E_OIDCNT  -20 */  "invalid number of outlier indicators\n",
  /* E_RFNPAR  -21 */  "invalid radial function parameter %g",
  /* E_NRMPAR  -22 */  "invalid normalization parameter %g",
  /* E_MFEXP   -23 */  "invalid membership/weight exponent %g",
  /* E_METHOD  -24 */  "unknown parameter update method %s",
  /* E_MODIFY  -25 */  "unknown parameter update modifier %s",
  /* E_MOMENT  -26 */  "invalid momentum coefficient %g",
  /* E_RADIUS  -27 */  "invalid initial radius %g",
  /* E_REGPAR  -28 */  "invalid regularization parameter %g",
  /* E_LRATE   -29 */  "invalid learning rate %g",
  /* E_UPDPAR  -30 */  "invalid update parameter %g",
  /* E_EPOCHS  -31 */  "invalid number of epochs %"DIMID_FMT,
  /*           -32 */  "unknown error"
};

static const MODEINFO initab[] = { /* table of initialization modes */
  { CLS_CENTER,    "center",
                   "center of the data space"                         },
  { CLS_CENTER,    "ctr",                                        NULL },
  { CLS_CENTER,    "c",                                          NULL },
  { CLS_UNIFORM,   "uniform",
                   "samples from a uniform distribution"              },
  { CLS_UNIFORM,   "unif",                                       NULL },
  { CLS_UNIFORM,   "uni",                                        NULL },
  { CLS_UNIFORM,   "u",                                          NULL },
  { CLS_DIAG,      "diag",
                   "points on the diagonal of the data space"         },
  { CLS_DIAG,      "d",                                          NULL },
  { CLS_LATIN,     "latin",
                   "latin hypercube sampling"                         },
  { CLS_LATIN,     "lhcs",                                       NULL },
  { CLS_LATIN,     "lhc",                                        NULL },
  { CLS_LATIN,     "lhs",                                        NULL },
  { CLS_LATIN,     "l",                                          NULL },
  { CLS_POINTS,    "points",
                   "data points chosen uniformly at random"           },
  { CLS_POINTS,    "pts",                                        NULL },
  { CLS_POINTS,    "p",                                          NULL },
  { CLS_MAXIMIN,   "maximin",
                   "choice by maximum minimum distance"               },
  { CLS_MAXIMIN,   "maxmin",                                     NULL },
  { CLS_MAXIMIN,   "max",                                        NULL },
  { CLS_MAXIMIN,   "m",                                          NULL },
  { CLS_KMPP,      "kmeans++",
                   "sampling from d^2 distribution"                   },
  { CLS_KMPP,      "km++",                                       NULL },
  { CLS_KMPP,      "kmpp",                                       NULL },
  { CLS_KMPP,      "kmeanspp",                                   NULL },
  { CLS_KMPP,      "cmeans++",                                   NULL },
  { CLS_KMPP,      "kc++",                                       NULL },
  { CLS_KMPP,      "cmpp",                                       NULL },
  { CLS_KMPP,      "cmeanspp",                                   NULL },
  { CLS_KMPP,      "k",                                          NULL },
  { CLS_KMPP,      "+",                                          NULL },
  { CLS_KMPP,      "++",                                         NULL },
  { CLS_QKMPP,     "qkmeans++",
                   "sampling from quantile range of d^2 distribution" },
  { CLS_QKMPP,     "qkm++",                                      NULL },
  { CLS_QKMPP,     "qkmpp",                                      NULL },
  { CLS_QKMPP,     "qkmeanspp",                                  NULL },
  { CLS_QKMPP,     "q",                                          NULL },
  { CLS_AFKMC2,    "afkmc2",
                   "Markov chain based sampling from d^2"             },
  { CLS_AFKMC2,    "mc2",                                        NULL },
  { CLS_AFKMC2,    "a",                                          NULL },
  { CLS_SAFKMC2,   "safkmc2",
                   "simplified Markov chain based sampling"           },
  { CLS_SAFKMC2,   "smc2",                                       NULL },
  { CLS_SAFKMC2,   "s",                                          NULL },
  { CLS_RKMPP,     "rkmeans++",
                   "sampling from distance distribution"              },
  { CLS_RKMPP,     "rkm++",                                      NULL },
  { CLS_RKMPP,     "rkmpp",                                      NULL },
  { CLS_RKMPP,     "rkmeanspp",                                  NULL },
  { CLS_RKMPP,     "rcmeans++",                                  NULL },
  { CLS_RKMPP,     "rkc++",                                      NULL },
  { CLS_RKMPP,     "rcmpp",                                      NULL },
  { CLS_RKMPP,     "rcmeanspp",                                  NULL },
  { CLS_RKMPP,     "rk",                                         NULL },
  { CLS_RKMPP,     "r+",                                         NULL },
  { CLS_RKMPP,     "r++",                                        NULL },
  { CLS_RQKMPP,    "rqkmeans++",
                   "sampling from quantile range of d distribution"   },
  { CLS_RQKMPP,    "rqkm++",                                     NULL },
  { CLS_RQKMPP,    "rqkmpp",                                     NULL },
  { CLS_RQKMPP,    "rqkmeanspp",                                 NULL },
  { CLS_RQKMPP,    "rq",                                         NULL },
  { CLS_RAFKMC2,   "rafkmc2",
                   "Markov chain based sampling from distance"        },
  { CLS_RAFKMC2,   "rmc2",                                       NULL },
  { CLS_RAFKMC2,   "ra",                                         NULL },
  { CLS_RSAFKMC2,  "rsafkmc2",
                   "simplified Markov chain based sampling"           },
  { CLS_RSAFKMC2,  "rsmc2",                                      NULL },
  { CLS_RSAFKMC2,  "rs",                                         NULL },
  { -1,            NULL,        /* sentinel */                   NULL },
};

static const MODEINFO nrmtab[] = { /* table of normalization modes */
  { CLS_NONE,      "none",      "no normalization"                    },
  { CLS_SUM1,      "sum1",      "normalize to sum 1"                  },
  { CLS_MAX1,      "max1",      "normalize to maximum 1"              },
  { CLS_HARD,      "hard",      "hard assignment"                     },
  { -1,            NULL,        /* sentinel */                   NULL },
};

static const MODEINFO updtab[] = { /* table of update methods */
/*{ CLS_GRADIENT,  "gradient",  "gradient based method"            }, */
  { CLS_ALTOPT,    "altopt",    "alternating optimization"            },
  { CLS_ALTOPT,    "ao",                                         NULL },
  { CLS_COMPLRN,   "complrn",   "competitive learning"                },
  { CLS_COMPLRN,   "cl",                                         NULL },
  { CLS_STANDARD,  "standard",  "standard Lloyd method *"             },
  { CLS_STANDARD,  "std",                                        NULL },
  { CLS_ELKAN,     "elkan",     "method from [Elkan 2003] *"          },
  { CLS_ELKAN,     "elk",                                        NULL },
  { CLS_HAMERLY,   "hamerly",   "method from [Hamerly 2010] *"        },
  { CLS_HAMERLY,   "ham",                                        NULL },
  { CLS_EXPONION,  "exponion",  "equiv. to -Ahamerly -ac2call *"      },
  { CLS_EXPONION,  "expo",                                       NULL },
  { CLS_EXPONION,  "exp",                                        NULL },
  { CLS_EXPONION,  "exon",                                       NULL },
  { CLS_EXPONION,  "onion",                                      NULL },
  { CLS_SHALLOT,   "shallot",   "shallot algorithm [Borgelt 2020] *"  },
  { CLS_SHALLOT,   "sha",                                        NULL },
  { CLS_SHALLOT,   "sht",                                        NULL },
  { CLS_ANNULAR,   "annular",   "method from [Drake 2013] *"          },
  { CLS_ANNULAR,   "annulus",                                    NULL },
  { CLS_ANNULAR,   "annuli",                                     NULL },
  { CLS_ANNULAR,   "ann",                                        NULL },
  { CLS_STRATUM,   "stratum",   "strata instead of annuli *"          },
  { CLS_STRATUM,   "strata",                                     NULL },
  { CLS_STRATUM,   "strat",                                      NULL },
  { CLS_STRATUM,   "str",                                        NULL },
  { CLS_COGRID,    "cogrid",    "coordinate grid algorithm *"         },
  { CLS_COGRID,    "cog",                                        NULL },
  { CLS_COGRID,    "grid",                                       NULL },
  { CLS_COGRID,    "cg",                                         NULL },
  { CLS_HYBRID,    "hybrid",    "hybrid of Hamerly and Elkan *"       },
  { CLS_HYBRID,    "hyb",                                        NULL },
  { CLS_HYSHALL,   "hyshall",   "hybrid with shallot algorithm *"     },
  { CLS_HYSHALL,   "hysha",                                      NULL },
  { CLS_HYSHALL,   "hysht",                                      NULL },
  { CLS_HYSHALL,   "hshallot",                                   NULL },
  { CLS_HYSHALL,   "hsha",                                       NULL },
  { CLS_HYSHALL,   "hsht",                                       NULL },
  { CLS_HYSHALL,   "hsh",                                        NULL },
  { CLS_HYSHALL,   "hs",                                         NULL },
  { CLS_YINYANG,   "yinyang",   "method from [Ding et al. 2015] *"    },
  { CLS_YINYANG,   "yin",                                        NULL },
  { CLS_YINYANG,   "yy",                                         NULL },
  { -1,            NULL,        /* sentinel */                   NULL },
};

static const MODEINFO modtab[] = { /* table of update modifiers */
  { CLS_NONE,      "none",      "standard optimization"               },
  { CLS_EXPAND,    "expand",    "expand change by a factor"           },
  { CLS_MOMENTUM,  "momentum",  "update with momentum term"           },
  { CLS_ADAPTIVE,  "adaptive",  "self-adaptive change factor"         },
  { CLS_RESILIENT, "resilient", "resilient update"                    },
  { CLS_QUICK,     "quick",     "quickprop analog"                    },
  { CLS_C2CMIN,    "c2cmin",
                   "cluster to cluster distances (only minima) *"     },
  { CLS_C2CMIN,    "c2cm",                                       NULL },
  { CLS_C2CMIN,    "cmin",                                       NULL },
  { CLS_C2CALL,    "c2call",
                   "cluster to cluster distances (all) *"             },
  { CLS_C2CALL,    "c2ca",                                       NULL },
  { CLS_C2CALL,    "c2c",                                        NULL },
  { CLS_C2CALL,    "call",                                       NULL },
  { -1,            NULL,        /* sentinel */                   NULL },
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner (for domain descriptions) */
static ATTSET  *attset = NULL;  /* attribute set */
static ATTMAP  *attmap = NULL;  /* attribute map */
static TABREAD *tread  = NULL;  /* table reader */
static TABLE   *table  = NULL;  /* training data points as a table */
static MATRIX  *data   = NULL;  /* training data points as a matrix */
static MATRIX  *ctrs   = NULL;  /* cluster centers for hard c-means */
static CLSET   *clset  = NULL;  /* cluster set for other clustering */
static NSTATS  *nst    = NULL;  /* numeric statistics for scaling */
static DIMID   *outids = NULL;  /* outlier identifiers/indices */
static FILE    *out    = NULL;  /* output file */
static char    *buf    = NULL;  /* buffer for file names */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (ctrs)   mat_delete(ctrs);     \
  if (clset)  cls_delete(clset);    \
  if (data)   mat_delete(data);     \
  if (nst)    nst_delete(nst);      \
  if (table)  tab_delete(table, 0); \
  if (attmap) am_delete(attmap, 0); \
  if (attset) as_delete(attset);    \
  if (tread)  trd_delete(tread, 1); \
  if (scan)   scn_delete(scan,  1); \
  if (outids) free(outids);         \
  if (buf)    free(buf);            \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void help (void)
{                               /* --- print help on modes */
  int i;                        /* loop variable */

  fprintf(stderr, "\n");        /* terminate startup message */
  printf("list of initialization modes (option -i#):\n");
  printf("(choice of initial positions of the cluster centers;\n");
  printf("a random offset can be added with the option -o#)\n");
  printf("  name        initialization mode\n");
  for (i = 0; initab[i].name; i++)
    if (initab[i].desc)         /* list of initialization modes */
      printf("  %-10s  %s\n", initab[i].name, initab[i].desc);
  printf("list of membership normalization modes (option -E#)\n");
  printf("  name        membership normalization mode\n");
  for (i = 0; nrmtab[i].name; i++)
    if (nrmtab[i].desc)         /* list of normalization modes */
      printf("  %-10s  %s\n", nrmtab[i].name, nrmtab[i].desc);
  printf("list of parameter update methods (option -A#)\n");
  printf("  name        parameter update method\n");
  for (i = 0; updtab[i].name; i++)
    if (updtab[i].desc)         /* list of update methods */
      printf("  %-10s  %s\n", updtab[i].name, updtab[i].desc);
  printf("marked with *: exact hard c-means methods (only centers)\n");
  printf("list of parameter update modifiers (option -a#)\n");
  printf("  name        parameter update method\n");
  for (i = 0; modtab[i].name; i++)
    if (modtab[i].desc)         /* list of update modifiers */
      printf("  %-10s  %s\n", modtab[i].name, modtab[i].desc);
  printf("marked with *: exact hard c-means methods (only centers)\n");
  exit(0);                      /* abort the program */
}  /* help() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter, buffer */
  char    *s, *end;             /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input cluster file */
  CCHAR   *fn_dom  = NULL;      /* name of domain file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_out  = NULL;      /* name of outlier indicator file */
  CCHAR   *fn_cls  = NULL;      /* name of output cluster file */
  CCHAR   *fname   = NULL;      /* name of output cluster file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *trgname = NULL;      /* name of optional target attribute */
  CCHAR   *ininame = "points";  /* name of initialization mode */
  CCHAR   *nrmname = "sum1";    /* name of normalization mode */
  CCHAR   *updname = "altopt";  /* name of parameter update method */
  CCHAR   *modname = "none";    /* name of parameter update modifier */
  int     matinp   = 0;         /* flag for numerical matrix input */
  int     mode     = AS_ATT|AS_NOXATT|AS_NONULL;/* table read mode */
  DIMID   clscnt   = 2;         /* number of clusters */
  int     type     = CLS_CENTER;/* cluster type */
  int     gauss    = 0;         /* flag for Gaussian function */
  double  rfnps[]  = {2, 0};    /* radial function parameters */
  int     nrmmode  = 0;         /* normalization mode */
  double  nrmps[]  = {1, 0};    /* normalization parameters */
  double  noise    = 0;         /* m.s. degree to noise cluster */
  int     owrite   = 0;         /* overwrite parameters with options */
  int     norm     = 1;         /* flag for input range normalization */
  int     inimode  = 0;         /* initialization mode */
  double  size     = 100;       /* init.: section size/chain length */
  double  trim     = 0;         /* init.: largest distance to trim */
  double  thout    = 0.5;       /* outlier indicator threshold */
  double  gcnt     = 0;         /* update parameter (groups/bounds) */
  double  range    = 0;         /* range for random offsets */
  double  radius   = 1.0;       /* initial cluster size (radius) */
  int     method   = CLS_MODVAR;/* parameter update method */
  DIMID   epochs   = 1000, e;   /* number of update epochs */
  DIMID   conly    = 0;         /* epochs to update centers only */
  DIMID   update   = 0,    u;   /* number of points between updates */
  int     shuffle  = 1;         /* shuffle data set */
  double  trmchg   = DBL_EPSILON;   /* maximum change for termination */
  double  msexp    = 2.0;       /* membership exponent for adaptation */
  double  moment   = 0;         /* momentum term coefficient */
  double  shrink   = 0.7;       /* shrink factor for learning rate */
  double  growth   = 1.2;       /* growth factor for learning rate */
  double  maxchg   = 1.8;       /* maximal change (neural update) */
  double  rates[]  = {0.2,0.2,0.2}; /* learning rates */
  double  decay[]  = {-.2,-.2,-.2}; /* learning rate decay params. */
  double  regps[]  = {0,1,1,0,0};   /* regularization parameters */
  double  fwexp    = 0;         /* exponent for feature weighting */
  int     sort     = 1;         /* whether to sort cluster by center */
  int     maxlen   = 0;         /* maximal output line length */
  DIMID   exclude  = 0;         /* number of target fields to exclude */
  int     report   = 0;         /* flag for intermediate reporting */
  long    seed;                 /* seed for random numbers */
  double  chg, x;               /* (max.) change of a center coord. */
  double  *doids;               /* outlier indicators as double */
  ATTID   a, c;                 /* number of attributes */
  DIMID   m;                    /* number of inputs/dimensions */
  DIMID   n, r;                 /* number of data points/vectors */
  DIMID   o = 0;                /* number of outlier tuples/vectors */
  DIMID   q;                    /* tuple identifier/index */
  double  w;                    /* weight of data points/vectors */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] [-M|domfile [-d|-h hdrfile]] "
                    "tabfile outfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-!       print a list of init. and norm. modes "
                    "and update methods\n");
    printf("-M       input is a numerical matrix            "
                    "(default: input is a table)\n");
    printf("-c#      number of clusters                     "
                    "(default: %"DIMID_FMT")\n", clscnt);
    printf("-v       adaptable variances                    "
                    "(default: unit variances)\n");
    printf("-V       adaptable covariances                  "
                    "(default: unit matrix)\n");
    printf("-Z       adaptable sizes                        "
                    "(default: fixed sizes)\n");
    printf("-j       joint size/(co)variances               "
                    "(default: individual)\n");
    printf("-w       adaptable weights/prior probs.         "
                    "(default: equal weights)\n");
    printf("-G       Gaussian radial function               "
                    "(default: Cauchy function)\n");
    printf("-p#:#    radial function parameters             "
                    "(default: %g:%g)\n", rfnps[0], rfnps[1]);
    printf("-N       normalize to unit integral             "
                    "(default: no normalization)\n");
    printf("-y#      membership to noise cluster            "
                    "(default: %g)\n", noise);
    printf("-J#:#    membership transform parameters        "
                    "(default: %g:%g)\n", nrmps[0], nrmps[1]);
    printf("-E#      membership normalization mode          "
                    "(default: %s)\n", nrmname);
    printf("-F#      exponent for feature weighting         "
                    "(default: %g)\n", fwexp);
    printf("-x#      exponent for data point weight         "
                    "(default: %g)\n", msexp);
    printf("-O       overwrite with option values           "
                    "(default: parameters read)\n");
    printf("-1       normalize centers to unit length       "
                    "(default: no normalization)\n");
    printf("-i#      initialization mode                    "
                    "(default: %s:%g:%g)\n", ininame, size, trim);
    printf("-L#      name of file with outlier indicators   "
                    "(default: none)\n");
    printf("-u#      threshold for outlier probability      "
                    "(default: %g)\n", thout);
    printf("-o#      random offset range                    "
                    "(default: %g)\n", range);
    printf("-S#      seed for random numbers                "
                    "(default: time)\n");
    printf("-I#      initial cluster size (radius)          "
                    "(default: %g)\n", radius);
    printf("-e#      maximum number of update epochs        "
                    "(default: %"DIMID_FMT")\n", epochs);
    printf("-Y#      epochs to update centers only          "
                    "(default: %"DIMID_FMT")\n", conly);
    printf("-k#      data points between two updates        "
                    "(default: %"DIMID_FMT")\n", update);
    printf("-s       do not shuffle data points             "
                    "(default: once per epoch)\n");
    printf("-q       do not normalize input ranges          "
                    "(default: mean=0, var=1)\n");
    printf("-A#      parameter update method                "
                    "(default: %s)\n", updname);
    printf("-a#      parameter update modifier              "
                    "(default: %s)\n", modname);
    printf("-K       use modifier only for centers          "
                    "(default: all parameters)\n");
    printf("-t#:#:#  learning rates                         "
                    "(default: %g:%g:%g)\n",
                     rates[0], rates[1], rates[2]);
    printf("-D#:#:#  learning rates decay parameters        "
                    "(default: %g:%g:%g)\n",
                     decay[0], decay[1], decay[2]);
    printf("-m#      momentum coefficient                   "
                    "(default: %g)\n", moment);
    printf("-g#:#    shrink and growth factor               "
                    "(default: %g:%g)\n", shrink, growth);
    printf("-z#      maximal change/learning rate           "
                    "(default: %g)\n", maxchg);
    printf("-T#      maximum change for termination         "
                    "(default: %g)\n", trmchg);
    printf("-H#      shape regularization parameter         "
                    "(default: %g)\n", regps[3]);
    printf("-R#:#:#  size  regularization parameters        "
                    "(default: %g:%g:%g)\n",
                     regps[0], regps[1], regps[2]);
    printf("-W#      weight regularization parameter        "
                    "(default: %g)\n", regps[4]);
    printf("-X#      exclude a target attribute             "
                    "(default: use all atts.)\n");
    printf("-P       report cluster state in each epoch     "
                    "(default: only final)\n");
    printf("-B       do not sort clusters by their centers  "
                    "(default: sort)\n");
    printf("-l#      output line length                     "
                    "(default: no limit)\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("domfile  file containing domain descriptions\n"
           "         (and optionally an initial cluster set)\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write induced cluster set to\n");
    printf("infile   file to read initial cluster set from "
                    "(only with -M)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */
  /* remaining option characters: u Q U */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* get a default seed value */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case '!': help();                              break;
          case 'M': matinp  = 1;                         break;
          case 'c': clscnt  = (DIMID)strtol(s, &s, 0);   break;
          case 'v': type   |= CLS_VARS;                  break;
          case 'V': type   |= CLS_COVARS;                break;
          case 'Z': type   |= CLS_SIZE;                  break;
          case 'j': type   |= CLS_JOINT;                 break;
          case 'w': type   |= CLS_WEIGHT;                break;
          case 'G': gauss   = 1;                         break;
          case 'p': getdblvec(s, &s, 2, rfnps);          break;
          case 'N': type   |= CLS_NORM;                  break;
          case 'y': noise   =      strtod(s, &s);        break;
          case 'J': getdblvec(s, &s, 2, nrmps);          break;
          case 'E': optarg  = &nrmname;                  break;
          case 'F': fwexp   =      strtod(s, &s);        break;
          case 'O': owrite  = 1;                         break;
          case '1': method |= CLS_UNIT;                  break;
          case 'x': msexp   =      strtod(s, &s);        break;
          case 'i': optarg  = &ininame;                  break;
          case 'L': optarg  = &fn_out;                   break;
          case 'u': thout   =      strtod(s, &s);        break;
          case 'o': range   =      strtod(s, &s);        break;
          case 'S': seed    =      strtol(s, &s, 0);     break;
          case 'I': radius  =      strtod(s, &s);        break;
          case 'e': epochs  = (int)strtol(s, &s, 0);     break;
          case 'Y': conly   = (int)strtol(s, &s, 0);     break;
          case 'k': update  = (int)strtol(s, &s, 0);     break;
          case 's': shuffle = 0;                         break;
          case 'q': norm    = 0;                         break;
          case 'A': optarg  = &updname;                  break;
          case 'a': optarg  = &modname;                  break;
          case 'K': method &= ~CLS_MODVAR;               break;
          case 't': getdblvec(s, &s, 3, rates);          break;
          case 'D': getdblvec(s, &s, 3, decay);          break;
          case 'm': moment  =      strtod(s, &s);        break;
          case 'g': getdbls(s, &s, 2, &shrink, &growth); break;
          case 'z': maxchg  =      strtod(s, &s);        break;
          case 'T': trmchg  =      strtod(s, &s);        break;
          case 'H': getdbls(s, &s, 1, regps+3);          break;
          case 'R': getdblvec(s, &s, 3, regps);          break;
          case 'W': getdbls(s, &s, 1, regps+4);          break;
          case 'X': optarg  = &trgname;                  break;
          case 'P': report  = 1;                         break;
          case 'B': sort    = 0;                         break;
          case 'l': maxlen  = (int)strtol(s, &s, 0);     break;
          case 'r': optarg  = &recseps;                  break;
          case 'f': optarg  = &fldseps;                  break;
          case 'b': optarg  = &blanks;                   break;
          case 'C': optarg  = &comment;                  break;
          case 'n': mode   |= AS_WEIGHT;                 break;
          case 'd': mode   |= AS_DFLT;                   break;
          case 'h': optarg  = &fn_hdr;                   break;
          default : error(E_OPTION, *--s);               break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_cls = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if (matinp) {                 /* if matrix version */
    if ((k != 2) && (k != 3))   /* check the number */
      error(E_ARGCNT);          /* of arguments */
    fn_inp = fn_cls;            /* shift the file names */
    fn_cls = fn_tab;            /* (as there is no domain file) */
    fn_tab = fn_dom; fn_dom = NULL;
    if ((!fn_tab || !*fn_tab) && (!fn_inp || !*fn_inp))
      error(E_STDIN); }         /* stdin must not be used twice */
  else {                        /* if table version */
    if (k != 3) error(E_ARGCNT);/* check the number of arguments */
    if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
      fn_hdr = "";              /* convert "-" to "" */
    i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
    if  (!fn_dom || !*fn_dom) i++;
    if  (!fn_tab || !*fn_tab) i++;
    if (i > 1) error(E_STDIN);  /* check assignments of stdin: */
  }                             /* stdin must not be used twice */
  if (rfnps[0] <= 0) error(E_RFNPAR, rfnps[0]);
  if (rfnps[1] <  0) error(E_RFNPAR, rfnps[1]);
  if (radius   <= 0) error(E_RADIUS, radius);
  s = strchr(ininame, ':');     /* check for a init. parameter */
  if (s != NULL) {              /* if such a parameter is given, */
    *s++ = 0;                   /* split into name and parameter */
    size = strtod(s, &end);     /* and get the parameter values */
    if (*end == ':') trim = strtod(s = end+1, &end);
    if (*end) error(E_INIPAR, s);
  }                             /* check for trailing info. */
  for (i = 0; initab[i].name; i++)
    if (strcmp(initab[i].name, ininame) == 0) break;
  if (!initab[i].name) error(E_MODE,   ininame);
  inimode = initab[i].code;    /* code the initialization method */
  if (size > 0) size /= 100;   /* turn percentages into fractions */
  if (trim > 0) trim /= 100;
  for (i = 0; nrmtab[i].name; i++)
    if (strcmp(nrmtab[i].name, nrmname) == 0) break;
  if (!nrmtab[i].name) error(E_MODE,   nrmname);
  nrmmode = nrmtab[i].code;     /* code the normalization method */
  if (nrmps[0]  <= -1) error(E_NRMPAR, nrmps[0]);
  if (nrmps[1]  <   0) error(E_NRMPAR, nrmps[1]);
  if ((nrmps[0] ==  0) && (msexp == 0)) error(E_NRMPAR, nrmps[0]);
  s = strchr(updname, ':');     /* check for an update parameter */
  if (s != NULL) {              /* if such a parameter is given, */
    *s++ = 0;                   /* split into name and parameter */
    gcnt = strtod(s, &end);     /* and get the parameter values */
    if (*end) error(E_UPDPAR, s);
  }                             /* check for trailing info. */
  for (i = 0; updtab[i].name; i++)
    if (strcmp(updtab[i].name, updname) == 0) break;
  if (!updtab[i].name) error(E_METHOD, updname);
  method |= updtab[i].code;     /* code the update/optim. method */
  if (gcnt > 0) gcnt /= 100;    /* turn percentage into fraction */
  for (i = 0; modtab[i].name; i++)
    if (strcmp(modtab[i].name, modname) == 0) break;
  if (!modtab[i].name) error(E_MODIFY, modname);
  i = modtab[i].code;           /* code the update modification */
  if (i != 0) method |= i;      /* combine method and modifier */
  else        method &= ~CLS_MODVAR;
  if (method & CLS_PLAIN) { type = CLS_CENTER; nrmmode = CLS_HARD; }
  if (growth < shrink) { w = growth; growth = shrink; shrink = w; }
  if (msexp    < 0) error(E_MFEXP,  msexp);
  if (fwexp    < 0) error(E_MFEXP,  fwexp);
  if (moment   < 0) error(E_MOMENT, moment);
  if (maxchg   < 1) error(E_UPDPAR, maxchg);
  if (growth   < 1) error(E_UPDPAR, growth);
  if (shrink   < 0) error(E_UPDPAR, shrink);
  if (rates[0] < 0) error(E_LRATE,  rates[0]);
  if (rates[1] < 0) error(E_LRATE,  rates[1]);
  if (rates[2] < 0) error(E_LRATE,  rates[2]);
  if (epochs   < 0) error(E_EPOCHS, epochs);
  rseed((unsigned)seed);        /* init. the random number generator */
  if (!(fn_cls && *fn_cls)) report = 0;
  if (report) {                 /* if to report intermediate results */
    buf = malloc((strlen(fn_cls) +20) *sizeof(char));
    if (!buf) error(E_NOMEM);   /* create a buffer for the file names */
  }                             /* that have to be constructed */
  fputc('\n', stderr);          /* terminate the startup message */

  if (matinp) {                 /* if matrix version */
    /* --- parse cluster descriptions --- */
    if (!fn_inp) m = -1;        /* if no input, clear att. counter */
    else {                      /* if an input file is given */
      t    = clock();           /* start the timer */
      scan = scn_create();      /* create a scanner */
      if (!scan) error(E_NOMEM);/* for the input cluster set */
      if (scn_open(scan, NULL, fn_inp) != 0)
        error(E_FOPEN, scn_name(scan));
      fprintf(stderr, "reading %s ... ", scn_name(scan));
      if (method & CLS_PLAIN) { /* if plain (center) update method */
        ctrs = cls_cparse(scan, 0, 1, &nst);
        if (!ctrs || !scn_eof(scan, 1))   /* parse centers and */
          error(E_PARSE, scn_name(scan)); /* scaling parameters */
        m      = mat_colcnt(ctrs);        /* get number of dimensions */
        clscnt = mat_rowcnt(ctrs);        /* and number of clusters */
        if (!nst) norm = 0; }   /* check input scales */
      else {                    /* if extended (fuzzy) update method */
        clset = cls_parse(scan, 0, 1);    /* parse cluster descs. */
        if (!clset || !scn_eof(scan, 1))  /* as a full cluster set */
          error(E_PARSE, scn_name(scan)); /* (not just centers) */
        m      = cls_inpcnt(clset);       /* get number of dimensions */
        clscnt = cls_clscnt(clset);       /* and number of clusters */
      }
      if (m < 1) error(E_INPCNT); /* check for at least one input */
      fprintf(stderr, "[%"DIMID_FMT" input(s),",   m);
      fprintf(stderr, " %"DIMID_FMT" cluster(s)]", clscnt);
      scn_delete(scan, 1);      /* delete the scanner */
      scan = NULL;              /* and clear the variable */
      fprintf(stderr, " done [%"CLK_FMT"s].\n", SEC_SINCE(t));
    } }                         /* print a success message */

  else {                        /* if table version */
    /* --- parse domain descriptions --- */
    attset = as_create("domains", att_delete);
    if (!attset) error(E_NOMEM);/* create an attribute set */
    scan = scn_create();        /* create a scanner */
    if (!scan)   error(E_NOMEM);/* for the domain file */
    t = clock();                /* start timer, open input file */
    if (scn_open(scan, NULL, fn_dom) != 0)
      error(E_FOPEN, scn_name(scan));
    fprintf(stderr, "reading %s ... ", scn_name(scan));
    if (as_parse(attset, scan, AT_ALL, 1) != 0)
      error(E_PARSE, scn_name(scan));
    a = as_attcnt(attset);      /* parse domain descriptions */
    if (a < 1) error(E_ATTCNT); /* check for at least one attribute */

    /* --- get optional target attribute --- */
    if (!scn_eof(scan, 0)) trgname = NULL;
    c = as_target(attset, trgname, -1);
    if (c < 0) error(trgname ? E_UNKTRG : E_MULTRG, trgname);
    attmap = am_create(attset, 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map */
    if (c < a)                  /* and set the target attribute */
      am_target(attmap, c);     /* (prevent it from being used) */
    m = am_inpcnt(attmap);      /* get the number of inputs */

    /* --- parse cluster set --- */
    if (scn_eof(scan, 0))       /* if there is no cluster set */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", a);
    else {                      /* if there is a cluster set, */
      mode |= AS_NOXVAL;        /* prevent nominal domain extensions */
      if (method & CLS_PLAIN) { /* if plain (center) update method */
        ctrs = cls_cparse(scan, m, 1, &nst);
        if (!ctrs || !scn_eof(scan, 1))   /* parse centers and */
          error(E_PARSE, scn_name(scan)); /* scaling parameters */
        clscnt = mat_rowcnt(ctrs);        /* get number of clusters */
        if (!nst) norm = 0; }   /* check input scales */
      else {                    /* if extended update method */
        clset = cls_parse(scan, m, 1);    /* parse cluster descs. */
        if (!clset || !scn_eof(scan, 1))
          error(E_PARSE, scn_name(scan));
        clscnt = cls_clscnt(clset);       /* get number of clusters */
      }
      fprintf(stderr, "[%"ATTID_FMT" input(s),",   m);
      fprintf(stderr, " %"DIMID_FMT" cluster(s)]", clscnt);
    }                           /* print a success message */
    scn_delete(scan, 1);        /* delete the scanner and */
    scan = NULL;                /* clear the scanner variable */
    fprintf(stderr, " done [%"CLK_FMT"s].\n", SEC_SINCE(t));
  }  /* if (matinp) .. else .. */

  if (matinp) {                 /* if matrix version */
    /* --- read data points/vectors --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = mat_readx(&data, tread, -1, m);
    if (k) error(k, TRD_INFO(tread));
    trd_delete(tread, 1);       /* read the data points/vectors, */
    tread = NULL;               /* then close the input file */
    m = mat_colcnt(data);       /* get the number of dimensions */
    n = mat_rowcnt(data);       /* and the number of data points */
    w = (double)n;              /* assume unit row weights */
    if (trgname) {              /* if to exclude columns */
      exclude = (DIMID)strtol(trgname, &s, 0);
      if ((s == trgname) || *s || (exclude >= m))
        error(E_EXCNT, trgname);
    }                           /* get number of excluded columns */
    if (exclude > 0)            /* if not to use all columns */
      mat_shrink(data, 0, m -= exclude);
    fprintf(stderr, "[%"DIMID_FMT" variable(s),", m);
    fprintf(stderr, " %"DIMID_FMT" vector(s)]",   n);
    fprintf(stderr, " done [%"CLK_FMT"s].\n", SEC_SINCE(t));
    if (m < 1) error(E_INPCNT); /* check for at least one input */
    if (n < 1) error(E_TPLCNT); /* and   for at least one data point */
  }  /* if (matinp) .. */

  else {                        /* if table version */
    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* configure the characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", a);
      fprintf(stderr, " done [%"CLK_FMT"s].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- read table --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    table = tab_create("table", attset, tpl_delete);
    if (!table) error(E_NOMEM); /* create a data table */
    k = tab_read(table, tread, mode);
    if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
    trd_delete(tread, 1);       /* read the table body and */
    tread = NULL;               /* delete the table reader */
    a = tab_colcnt(table);      /* get the number of attributes */
    n = tab_tplcnt(table);      /* and the number of data tuples */
    w = tab_tplwgt(table);      /* and print a success message */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
    fprintf(stderr, " %"DIMID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%"CLK_FMT"s].\n", SEC_SINCE(t));
    if (n < 1) error(E_TPLCNT); /* check for at least one data point */

    /* --- convert into matrix --- */
    data = tab_matrix(table, attmap, AM_INPUTS, 1);
    if (!data) error(E_NOMEM);  /* convert the read table */
    table = NULL;               /* into a (data) matrix */
  }  /* if (matinp) .. else .. */

  /* --- read outlier indicators --- */
  if (fn_out) {                 /* if outlier indicators are given */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* configure the characters */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_out) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    doids = (double*)malloc((size_t)n *sizeof(double));
    if (!doids) error(E_NOMEM); /* create outlier indicators and */
    outids = (DIMID*)doids;     /* use same memory for identifiers */
    for (r = 0; 1; ) {          /* traverse the table/matrix rows */
      k = trd_read(tread);      /* read the next field (number) */
      if (k <= TRD_ERR) error(E_FREAD, trd_name(tread));
      if (k <= TRD_EOF) break;  /* check for error and end of file */
      s = trd_field(tread);     /* get the field content */
      if (!*s) continue;        /* skip empty fields */
      w = strtod(s, &end);      /* convert field to a number */
      if ((end == s) || *end) error(E_OUTIND, s);
      if (r >= n) error(E_OIDCNT);
      doids[r++] = w;           /* check the conversion and */
    }                           /* store the read indicator */
    trd_delete(tread, 1);       /* read the table body and */
    tread = NULL;               /* delete the table reader */
    fprintf(stderr, "[%"DIMID_FMT" value(s)]", r);
    fprintf(stderr, " done [%"CLK_FMT"s].\n", SEC_SINCE(t));
    if (r < n) {                /* if fewer indicators than tuples */
      for (o = 0; o < r; o++) { /* traverse tuple identifiers/indices */
        q = (DIMID)doids[o];    /* get tuple identifier/index */
        if (((double)q != doids[o]) || (q < 0) || (q >= n))
          error(E_OUTID, doids[o]);
        outids[o] = q;          /* check for valid index and */
      } }                       /* store the tuple identifier/index */
    else {                      /* if one indicator per tuple */
      for (r = o = 0; r < n; r++)
        if (doids[r] >= thout) outids[o++] = r;
    }                           /* collect indices >= threshold */
    /* At this point the variable o contains the number of outliers, */
    /* the tuple indices of which are stored in the arrays outids. */
  }  /* if (fn_out) .. */

  /* --- reduce data --- */
  if ((update <= 0)             /* if to update once each epoch */
  &&  !fn_out) {                /* and no outlier indicators given */
    fprintf(stderr, "reducing %s ... ", (matinp) ? "matrix" : "table");
    n = mat_reduce(data);       /* reduce matrix for speed up */
    fprintf(stderr, "[%"DIMID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, (matinp) ? " vector" : " tuple");
    fprintf(stderr, "(s)] done [%"CLK_FMT"s].\n", SEC_SINCE(t));
  }                             /* print a log message */

  if (method & CLS_PLAIN) {     /* if plain (center) update method */
    /* --- normalize data --- */
    if (norm) {                 /* if to z-normalize the data */
      t = clock();              /* start timer, print message */
      fprintf(stderr, "normalizing data ... ");
      if (!nst) {               /* if there are no scaling factors */
        nst = nst_create(m);    /* create simple numeric statistics */
        if (!nst) error(E_NOMEM); /* to compute mean/std. deviation */
        for (r = 0; r < n; r++) /* register data for statistics */
          nst_reg(nst, mat_row(data, r), mat_rgetwgt(data, r));
        nst_reg(nst, NULL, 0);  /* compute the numeric statistics */
      }                         /* (mean and standard deviation) */
      for (r = 0; r < n; r++)   /* z-normalize the data */
        nst_norm(nst, mat_row(data, r), mat_row(data, r));
      fprintf(stderr, "[%"DIMID_FMT, n);
      if (w != (double)n) fprintf(stderr, "/%g", w);
      fprintf(stderr, (matinp) ? " vector" : " tuple");
      fprintf(stderr, "(s)] done [%"CLK_FMT"s].\n", SEC_SINCE(t));
    }                           /* print a log message */

    /* --- initialize clusters --- */
    if (!ctrs) {                /* if no cluster centers were read */
      t = clock();              /* start timer, print message */
      fprintf(stderr, "initializing clusters ... ");
      ctrs = mat_createx(clscnt, m, MAT_ROWWGTS);
      if (!ctrs                 /* create cluster centers (matrix) */
      || cls_ctrsel(ctrs, data, inimode,size,trim, outids,o, drand) < 0)
        error(E_NOMEM);         /* initialize the cluster centers */
      if (range > 0)            /* jitter initial cluster centers */
        cls_ctrjtr(ctrs, range, drand);
      fprintf(stderr, "[%"DIMID_FMT" cluster(s)]", clscnt);
      fprintf(stderr, " done [%"CLK_FMT"s].", SEC_SINCE(t));
      #ifdef BENCH              /* if benchmark version */
      fprintf(stderr, " of: %.15g", cls_hobjfn(ctrs, data));
      #endif                    /* compute & print objective function */
      fprintf(stderr, "\n");    /* terminate the log message */
    }

    /* --- optimize clusters --- */
    t = clock();                /* start timer, print message */
    fprintf(stderr, "optimizing clusters ... ");
    e = cls_xhcm(ctrs, data, method, epochs, outids, o, gcnt, NULL);
    if (e < 0) error(E_NOMEM);  /* do exact hard c-means clustering */
    fprintf(stderr, "[%"DIMID_FMT" epoch(s)]", e);
    fprintf(stderr, " done [%"CLK_FMT"s].", SEC_SINCE(t));
    #ifdef BENCH                /* if benchmark version */
    fprintf(stderr, " of: %.15g", cls_hobjfn(ctrs, data));
    #endif                      /* compute & print objective function */
    fprintf(stderr, "\n");      /* terminate the log message */

    /* --- write clusters --- */
    t = clock();                /* start timer, open output file */
    if (fn_cls && (strcmp(fn_cls, "-") == 0)) fn_cls = "";
    if (fn_cls && *fn_cls) { out = fopen(fn_cls, "w"); }
    else                   { out = stdout; fn_cls = "<stdout>"; }
    fprintf(stderr, "writing %s ... ", fn_cls);
    if (!out) error(E_FOPEN, fn_cls);
    if (!matinp) {              /* if table version */
      if (as_desc(attset, out, AS_TITLE|AS_IVALS|AS_DIRS, maxlen) != 0)
        error(E_FWRITE, fname); /* describe attribute domains */
      fprintf(out, "\n");       /* leave one line empty */
    }
    if (cls_cdesc(ctrs, nst, out, CLS_TITLE|CLS_INFO, maxlen) != 0)
      error(E_FWRITE, fn_cls);  /* describe the cluster set */
    if (out && (((out == stdout) ? fflush(out) : fclose(out)) != 0))
      error(E_FWRITE, fn_cls);  /* close the output file and */
    out = NULL;                 /* print a success message */
    fprintf(stderr, "[%"ATTID_FMT" input(s),",   m);
    fprintf(stderr, " %"DIMID_FMT" cluster(s)]", clscnt);
    fprintf(stderr, " done [%"CLK_FMT"s].\n", SEC_SINCE(t));
  }  /* if (method & CLS_PLAIN) .. */

  else {                        /* if extended update method */
    /* --- initialize clusters --- */
    if (clset)                  /* if a input cluster set is given, */
      clscnt = cls_clscnt(clset);     /* get the number of clusters */
    else {                      /* if no input cluster set is given */
      t = clock();              /* start the timer, print message */
      fprintf(stderr, "initializing clusters ... ");
      clset = cls_create(m, clscnt);
      if (!clset) error(E_NOMEM);  /* create a cluster set */
      n = mat_rowcnt(data);     /* get the number of data points */
      for (r = 0; r < n; r++)   /* determine the ranges of values */
        cls_reg(clset, mat_row(data, r), 1);
      if (norm)                 /* if to normalize the data, */
        cls_reg(clset, NULL, 0);/* compute the scaling factors */
      cls_type(clset, type, radius *radius);
      if (inimode < CLS_POINTS) /* initialize the clusters */
        cls_init(clset, inimode, drand, NULL);
      else {                    /* or select initial centers */
        ctrs = mat_create(clscnt, m);
        if (!ctrs               /* create a matrix of cluster centers */
        || (cls_ctrsel(ctrs, data, inimode, size, trim,
                       outids, o, drand) < 0))
          error(E_NOMEM);       /* traverse the matrix rows */
        for (r = 0; r < clscnt; r++)
          cls_init(clset, CLS_POINTS, drand, mat_row(ctrs, r));
        mat_delete(ctrs); ctrs = NULL;
      }                         /* set each matrix row as a center */
      if (range > 0)            /* jitter initial cluster centers */
        cls_jitter(clset, range, drand);
      owrite = -1;              /* set cluster type and radius */
      fprintf(stderr, "[%"DIMID_FMT" cluster(s)]", clscnt);
      fprintf(stderr, " done [%"CLK_FMT"s].\n", SEC_SINCE(t));
    }                           /* print a success message */
    if (owrite) {               /* if to overwrite the parameters, */
      cls_type(clset, type,-1); /* set the cluster parameters */
      if (gauss) cls_radfn(clset, rf_gauss,  rfd_gauss,  rfnps);
      else       cls_radfn(clset, rf_cauchy, rfd_cauchy, rfnps);
      cls_norm (clset, nrmmode, nrmps);
      cls_noise(clset, noise);  /* (these parameters can be read */
      cls_fwexp(clset, fwexp);  /* from the cluster input file */
      cls_msexp(clset, msexp);  /* and may be replaced) */
    }
    cls_setup  (clset);         /* set up the cluster set and */
    cls_method (clset, method); /* set the parameter update method */
    cls_lrate  (clset, rates,  decay);  /* set the learning rates, */
    cls_moment (clset, moment);         /* the momentum coefficient, */
    cls_factors(clset, growth, shrink); /* the update factors and */
    cls_limits (clset, 1,      maxchg); /* the update limits */
    cls_regular(clset, regps);  /* set the regularization parameters */

    /* --- optimize clusters --- */
    t = clock();                /* start timer, print message */
    fprintf(stderr, "optimizing clusters ... ");
    if (update <= 0)            /* suppress unnecessary shuffling */
      shuffle = 0;              /* (batch training used anyway) */
    u = update;                 /* initialize the tuple counter */
    for (e = 0; e < epochs; e++) { /* traverse the epochs */
      if ((e & 0x00ff) == 0)    /* print the current number of epochs */
        fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e+1);
      if (report) {             /* if to report intermediate steps */
        sprintf(buf, "%s.%"DIMID_FMT, fn_cls, e);
        out = fopen(buf, "w");  /* open the output file */
        if (!out) error(E_FOPEN, buf);
        if (!matinp) {          /* if table version */
          k = AS_TITLE|AS_IVALS|AS_DIRS;
          if (as_desc(attset, out, k, maxlen) != 0)
            error(E_FWRITE, buf);
          fprintf(out, "\n");   /* describe attribute domains */
        }                       /* and leave one line empty */
        if (cls_desc(clset, out, CLS_TITLE|CLS_INFO, maxlen) != 0)
          error(E_FWRITE, buf); /* describe the cluster set and */
        i = fclose(out); out = NULL;   /* close the output file */
        if (i != 0) error(E_FWRITE, buf);
      }                         /* check for a write error */
      chg = 0;                  /* clear the maximal change */
      if (shuffle || (e == 0))  /* shuffle the data points/vectors */
        mat_shuffle(data, drand, 0, -1);  /* (at least in epoch 0) */
      for (r = mat_rowcnt(data); --r >= 0; ) {
        cls_aggr(clset, mat_row(data, r), 1);
        if ((update > 0) && (--u <= 0)) {
          u = update;           /* update every 'update' data points */
          x = cls_update(clset, (e < conly));
          if (x > chg) chg = x; /* determine the maximum change */
        }                       /* of a center coordinate */
      }
      if (update <= 0)          /* if no number of tuples is given, */
        chg = cls_update(clset, (e < conly));
      if ((e >= conly) && (chg <= trmchg)) {
        e++; break; }           /* if needed, update once per epoch */
    }                           /* check the termination criterion */
    fprintf(stderr, "[%"DIMID_FMT" cluster(s),", clscnt);
    fprintf(stderr, " %"DIMID_FMT" epoch(s)]", e);
    fprintf(stderr, " done [%"CLK_FMT"s].\n", SEC_SINCE(t));

    /* --- write clusters --- */
    fname = fn_cls;             /* get the output file name */
    if (report) {               /* construct final file name */
      sprintf(buf, "%s.%"DIMID_FMT, fn_cls, e); fname = buf; }
    t = clock();                /* start timer, open output file */
    if (fname && (strcmp(fname, "-") == 0)) fname = "";
    if (fname && *fname) { out = fopen(fname, "w"); }
    else                 { out = stdout; fname = "<stdout>"; }
    fprintf(stderr, "writing %s ... ", fname);
    if (!out) error(E_FOPEN, fname);
    if (!matinp) {              /* if table version */
      if (as_desc(attset, out, AS_TITLE|AS_IVALS|AS_DIRS, maxlen) != 0)
        error(E_FWRITE, fname); /* describe attribute domains */
      fprintf(out, "\n");       /* leave one line empty */
    }
    if (sort) cls_sort(clset);  /* sort clusters by their centers */
    if (cls_desc(clset, out, CLS_TITLE|CLS_INFO, maxlen) != 0)
      error(E_FWRITE, fname);   /* describe the cluster set */
    if (out && (((out == stdout) ? fflush(out) : fclose(out)) != 0))
      error(E_FWRITE, fname);   /* close the output file and */
    out = NULL;                 /* print a success message */
    fprintf(stderr, "[%"ATTID_FMT" input(s),",   cls_inpcnt(clset));
    fprintf(stderr, " %"DIMID_FMT" cluster(s)]", clscnt);
    fprintf(stderr, " done [%"CLK_FMT"s].\n", SEC_SINCE(t));
  }  /* if (method & CLS_PLAIN) .. else .. */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
