/*----------------------------------------------------------------------
  File    : cluster3.c
  Contents: cluster validity measures
  Author  : Christian Borgelt
  History : 2003.05.17 file created
            2008.04.08 bug in cls_evaggr fixed (application of msexp)
            2010.04.08 special handling of squares removed
----------------------------------------------------------------------*/
#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <assert.h>
#include "cluster.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define LN_2         0.69314718055994530942  /* ln(2) */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef double CLSVALFN (CLSET *clset);

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static double intexp (double x, DIMID n)
{                               /* --- fast integer exponentiation */
  double r = (n & 1) ? x : 1;   /* result */

  for (n >>= 1; n > 0; n >>= 1) {
    x *= x;                     /* traverse the powers of 2 and */
    if (n & 1) r *= x;          /* if the corr. exponent bit is set */
  }                             /* multiply them into the result */
  return r;                     /* return the result */
}  /* intexp() */

/*--------------------------------------------------------------------*/

static double dist (CLSET *clset, CLUSTER *c1, CLUSTER *c2)
{                               /* --- distance of clusters */
  DIMID  i;                     /* loop variable */
  double *b;                    /* center difference vector */

  for (b = clset->buf, i = clset->inpcnt; --i >= 0; )
    b[i] = c1->ctr[i] -c2->ctr[i];
  if (clset->type & CLS_COVARS) /* if to use covariances */
    return 0.5 *(mat_mulvmv(c1->inv, b) +mat_mulvmv(c2->inv, b));
  if (clset->type & CLS_VARS)   /* if to use only variances */
    return 0.5 *(mat_mulvdv(c1->inv, b) +mat_mulvdv(c2->inv, b));
  if (clset->type & CLS_SIZE)   /* if to use only isotropic variance */
    return vec_sqrlen(b, clset->inpcnt) / (0.5 *(c1->var +c2->var));
  return vec_sqrlen(b, clset->inpcnt);
}  /* dist() */                /* if to use no size */

/*----------------------------------------------------------------------
  Cluster Validity Functions
----------------------------------------------------------------------*/

static double objfn (CLSET *clset)
{                               /* --- objective function */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  r = 0;                /* result */

  for (c = clset->cls +(i = clset->clscnt); --i >= 0; )
    r += (--c)->val[1];         /* sum the weighted distance sums */
  return r;                     /* and return this sum */
}  /* objfn() */

/*--------------------------------------------------------------------*/

static double coverage (CLSET *clset)
{                               /* --- coverage of data points */
  return clset->eval.sum;       /* return the sum of membership degs. */
}  /* coverage() */

/*--------------------------------------------------------------------*/

static double fdbidx (CLSET *clset)
{                               /* --- fuzzy Davies-Bouldin index */
  DIMID   i, k;                 /* loop variables */
  CLUSTER *c1, *c2;             /* to traverse the clusters */
  double  max, t;               /* maximum of values, buffer */
  double  r = 0;                /* result */

  for (c1 = clset->cls +(i = clset->clscnt); --i >= 0; ) {
    --c1;                       /* traverse the clusters */
    c1->val[3] = (c1->val[0] > 0) ? c1->val[1]/c1->val[0] : 0;
  }                             /* compute the average distances */
  for (c1 = clset->cls +(i = clset->clscnt); --i >= 0; ) {
    --c1; max = 0;              /* traverse the clusters */
    for (c2 = clset->cls +(k = clset->clscnt); --k >= 0; ) {
      if (--c2 == c1) continue; /* traverse the other clusters */
      t = dist(clset, c1, c2);  /* compute the cluster distance */
      t = (t > 0) ? (c1->val[3] +c2->val[3]) /t : 0;
      if (t > max) max = t;     /* determine the maximum */
    }                           /* of the size to distance ratio */
    r += max;                   /* and sum these maxima */
  }
  return r /(double)clset->clscnt; /* return average over clusters */
}  /* fdbidx() */

/*--------------------------------------------------------------------*/

static double pcoeff (CLSET *clset)
{                               /* --- partition coefficient */
  return clset->eval.sqr /clset->eval.cnt;
}  /* pcoeff() */

/*--------------------------------------------------------------------*/

static double pcnorm (CLSET *clset)
{                               /* --- norm. partition coefficient */
  if ((clset->clscnt <= 1) || (clset->eval.cnt <= 0))
    return 0;                   /* check for valid denominators */
  return 1 -((double)clset->clscnt /(double)(clset->clscnt -1))
    *(1 -clset->eval.sqr /clset->eval.cnt);
}  /* pcnorm() */

/*--------------------------------------------------------------------*/

static double pentropy (CLSET *clset)
{                               /* --- partition entropy */
  return (clset->eval.ent < 0)
    ? -clset->eval.ent /((double)clset->clscnt*LN_2) : 0;
}  /* pentropy() */

/*--------------------------------------------------------------------*/

static double fhypvol (CLSET *clset)
{                               /* --- fuzzy hyper-volume */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  r = 0;                /* result */

  for (c = clset->cls +(i = clset->clscnt); --i >= 0; )
    r += pow((--c)->var, 0.5*(double)clset->inpcnt);
  return r;                     /* sum the roots of the determinants */
}  /* fhypvol() */              /* and return this sum */

/*--------------------------------------------------------------------*/

static double prpexp (CLSET *clset)
{                               /* --- proportion exponent */
  return clset->eval.pxp /LN_2;
}  /* prpexp() */

/*--------------------------------------------------------------------*/

static double xbisep (CLSET *clset)
{                               /* --- Xie-Beni index / separation */
  DIMID   i, k;                 /* loop variables */
  CLUSTER *c1, *c2;             /* to traverse the clusters */
  double  min = INFINITY;       /* minimum of cluster distances */
  double  r   = 0, t;           /* result, temporary buffer */

  if (clset->clscnt <= 1) return 0;
  for (c1 = clset->cls +(i = clset->clscnt); --i >= 0; ) {
    r += (--c1)->val[1];        /* compute the objective function */
    for (c2 = clset->cls +(k = clset->clscnt); --k >= 0; ) {
      if (--c2 == c1) continue; /* traverse the other clusters */
      t = dist(clset, c1, c2);  /* compute the cluster distance */
      if (t < min) min = t;     /* and determine the minimum */
    }                           /* of these distances */
  }
  min *= clset->eval.cnt;       /* compute the denominator and */
  return (min > 0) ? r/min : 0; /* return the separation degree */
}  /* xbisep() */

/*--------------------------------------------------------------------*/

static double partdens (CLSET *clset)
{                               /* --- partition density */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  r = 0, d = 0;         /* result and denominator */

  for (c = clset->cls +(i = clset->clscnt); --i >= 0; ) {
    r += (--c)->val[2];         /* sum the assigned points sum */
    d += pow(c->var, 0.5*(double)clset->inpcnt);
  }                             /* sum the roots of the determinants */
  return (d > 0) ? r/d : 0;     /* return the partition density */
}  /* partdens() */

/*--------------------------------------------------------------------*/

static double pdavg (CLSET *clset)
{                               /* --- average partition density */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  r = 0, t;             /* result, temporary buffer */

  for (c = clset->cls +(i = clset->clscnt); --i >= 0; ) {
    t = pow((--c)->var, 0.5*(double)clset->inpcnt);
    if (t > 0) r += c->val[2] /t;
  }                             /* cluster-spec. partition density */
  return r / (double)clset->clscnt;  /* compute and return average */
}  /* pdavg() */

/*--------------------------------------------------------------------*/

static double noise (CLSET *clset)
{                               /* --- sum of m.s. to noise cluster */
  return clset->eval.noise;
}  /* noise() */

/*--------------------------------------------------------------------*/

static CLSVALFN *clsvalfn[] = {
  /* CLS_OBJFN     0 */  objfn,
  /* CLS_COVERAGE  1 */  coverage,
  /* CLS_FDBIDX    2 */  fdbidx,
  /* CLS_PCOEFF    3 */  pcoeff,
  /* CLS_PCNORM    4 */  pcnorm,
  /* CLS_PENTROPY  5 */  pentropy,
  /* CLS_FHYPVOL   6 */  fhypvol,
  /* CLS_PRPEXP    7 */  prpexp,
  /* CLS_XBISEP    8 */  xbisep,
  /* CLS_PARTDENS  9 */  partdens,
  /* CLS_PDAVG    10 */  pdavg,
  /* CLS_NOISE    11 */  noise,
};

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

void cls_evcfg (CLSET *clset, int measure, double param)
{                               /* --- configure cluster evaluation */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  CLSEVAL *e = &clset->eval;    /* evaluation variables */

  assert(clset);                /* check the function argument */
  if (measure >= 0) {           /* if a measure is given, */
    e->flags |= 1 << measure;   /* set the measure flag and */
    if (param >= 0) e->rad = param; } /* note the parameter */
  else {                        /* if no measure is given, */
    e->flags = 0;               /* clear all measure flags */
    for (c = clset->cls +(i = clset->clscnt); --i >= 0; )
      c->val[0] = c->val[1] = c->val[2] = 0;
    e->cnt = e->sum = e->sqr = e->ent = e->pxp = e->noise = 0;
  }                             /* clear all parameters */
}  /* cls_evcfg() */

/*--------------------------------------------------------------------*/

void cls_evaggr (CLSET *clset, const double *vec, double weight)
{                               /* --- aggregate for evaluation */
  DIMID   i, k, n;              /* loop variables */
  CLUSTER *c;                   /* to traverse the clusters */
  CLSEVAL *e;                   /* evaluation variables */
  double  max = 0;              /* maximum membership degree */
  double  t, x, y, z;           /* temporary buffers */
  double  a = 0;                /* transformation parameter */

  assert(clset);                /* check the function argument */
  cls_exec(clset, vec, NULL);   /* execute the cluster set */
  e = &clset->eval;             /* get the evaluation variables */
  e->cnt += weight;             /* sum the pattern weight */
  x = clset->msexp;             /* get the adaptation exponent */
  if      (x == 0) x = 1;       /* and adapt it */
  else if (x <  0) a = (-x -1) /(-x +1);
  y = (e->flags & CLS_PENTROPY) ? 0 : INFINITY;
  for (c = clset->cls +(i = clset->clscnt); --i >= 0; ) {
    t = (--c)->msd;             /* traverse the clusters */
    e->sum += z = t*weight;     /* sum the membership degrees */
    e->sqr += t*z;              /* and their squares */
    if (t > max) max = t;       /* determine their maximum */
    if (t > y) e->ent += t *log(t);
    if (c->d2 < e->rad) c->val[2] += t;
    if (x < 0) t = (a *(t-1) +1) *t;
    else       t = pow(t, x);   /* use adaptation exponent */
    c->val[0] += t *= weight;   /* sum the membership degrees */
    c->val[1] += t *c->d2;      /* and the weighted distances */
  }                             /* for each cluster */
  e->noise += clset->ncmsd[1];  /* sum memberships to noise cluster */
  if (e->flags & CLS_PRPEXP) {  /* if to compute proportion exponent */
    x = 0; y = -1; i = clset->clscnt;
    n = (int)floor(1/max);      /* compute number of clusters */
    if (n > (clset->clscnt << 6)) n = clset->clscnt << 6;
    for (k = 1; k <= n; k++) {  /* traverse the clusters */
      y *= (double)(k -i -1) / (double)k;
      x += y *intexp(1 -(double)k *max, i -1);
    }                           /* add next term */
    e->pxp -= log(x);           /* sum the logarithms */
  }                             /* of cluster-specific terms */
}  /* cls_evaggr() */

/*--------------------------------------------------------------------*/

double cls_eval (CLSET *clset, int measure)
{                               /* --- compute evaluation measure */
  if ((measure < 0) || (measure >= CLS_VMCNT))
    return 0;                   /* check the measure range */
  return clsvalfn[measure](clset);
}  /* cls_eval() */
